// vim:ts=8:sw=8
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "main.h"
#include "zoom.h"


#ifndef DEBUG
#define fprintf(args...) do { }while(0)
#endif

int  	gWidth		= 0, 
	gHeight 	= 0,
	gZWidth 	= 0,
	gZHeight 	= 0,
	gByteLine	= 0,
	gDepth		= 0,
	gCurX		= 0,
	gCurY		= 0;

float gZoomFactor 	= DEFAULT_ZOOM;
int gUpdateRate 	= DEFAULT_UPDATE;
int gAutoUpdate 	= 0;
int gTimerId 		= 0;

int gFirstDrag 		= 1; // i still remember mine... *gasp*
int gIsDragging 	= 0;
int gIsSaved 		= 1;

GtkWidget *gWindow;
GtkWidget *gDrawing;
GtkWidget *gMenu;
GtkWidget *gFileSel;

GtkWidget *gZoomSpin;
GtkWidget *gUpdateSpin;

GdkVisual *gVisual;

static void update_dimension(int width, int height);
static void handle_expose(GtkWidget *wid, GdkEventExpose *event);


void handle_about(GtkWidget *widget, gpointer data)
{
}
static void handle_save_cancel(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(gFileSel);
}

static void handle_save_ok(GtkWidget *widget, gpointer data)
{
	char *fname;
	fname = gtk_file_selection_get_filename(GTK_FILE_SELECTION(gFileSel));
	if(fname != 0) {
		zoom_save(fname);
		gtk_widget_destroy(gFileSel);
	}
}

void
handle_save(GtkWidget *widget, gpointer data)
{
	gFileSel = gtk_file_selection_new("Save as PNG");
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(gFileSel)->ok_button),
			   "clicked", GTK_SIGNAL_FUNC(handle_save_ok), NULL);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(gFileSel)->cancel_button),
			   "clicked", GTK_SIGNAL_FUNC(handle_save_cancel), NULL);
	gtk_widget_show(gFileSel);
}

void
handle_save_as(GtkWidget *widget, gpointer data)
{
}

void
handle_quit(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}


static void 
set_zoomfactor(float f)
{
	gZoomFactor = f;
	
	update_dimension(gWidth, gHeight);
}

static void
handle_zoomfactor_cancel(GtkWidget *widget, gpointer data) 
{
	gtk_widget_destroy(widget);
}

static void
handle_zoomfactor_okay(GtkWidget *widget, gpointer data) 
{
	set_zoomfactor(gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(gZoomSpin)));
	handle_zoomfactor_cancel(widget, data);
}

static void
handle_zoomfactor_apply(GtkWidget *widget, gpointer data) 
{
	set_zoomfactor(gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(gZoomSpin)));
}

void
handle_zoomfactor(GtkWidget *widget, gpointer data)
{
	GtkWidget	*dialog,
			*okay,
			*cancel,
			*apply,
			*label;
	GtkAdjustment	*adj;
			
	dialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dialog), APPNAME": Zoom Factor" );
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);

	label = gtk_label_new("Default: "DEF_ZOOM_STR);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, 1,1,1);

	adj = (GtkAdjustment*)gtk_adjustment_new(gZoomFactor, 
			1.0, 120.0, 1.0, 1.0, 0.0);

	gZoomSpin = gtk_spin_button_new(adj, 1.0, 1);

	gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(gZoomSpin), FALSE);
	gtk_widget_set_usize(gZoomSpin, 100, 0);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), gZoomSpin, FALSE, TRUE, 0);


	okay = gtk_button_new_with_label("Okay");
	cancel = gtk_button_new_with_label("Cancel");
	apply = gtk_button_new_with_label("Apply");

	gtk_signal_connect_object(GTK_OBJECT(okay), "clicked", 
				  GTK_SIGNAL_FUNC(handle_zoomfactor_okay), 
				  (gpointer)dialog);
	gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", 
				  GTK_SIGNAL_FUNC(handle_zoomfactor_cancel), 
				  (gpointer)dialog);
	gtk_signal_connect_object(GTK_OBJECT(apply), "clicked", 
				  GTK_SIGNAL_FUNC(handle_zoomfactor_apply), 
				  (gpointer)dialog);

	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				cancel, 0,0,0); 
	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				okay, 0,0,0); 
	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				apply, 0,0,0); 



	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), gZoomSpin);
	gtk_widget_show_all (dialog);
}

static int update_cb(gpointer irrelevant)
{
	if(gAutoUpdate) {
//		fprintf(stderr, "timer update\n");
//		fprintf(stderr,"update_cb: gAutoUpdate=%d\n", gAutoUpdate);
		zoom_update_image();
	} else
		fprintf(stderr, "oops!\n");
	return TRUE;
}

void
handle_autoupdate(GtkWidget *widget, gpointer data)
{
	if(gAutoUpdate) {
//		fprintf(stderr, "handle_autoupdate: Removing timeout: %d\n",gTimerId);
		gtk_timeout_remove(gTimerId);
	} else {
		gTimerId = gtk_timeout_add(gUpdateRate, &update_cb, 0);
//		fprintf(stderr, "handle_autoupdate: Adding timeout: %d\n",gTimerId);
	}
	gAutoUpdate = !gAutoUpdate;
//	fprintf(stderr,"handle_autoupdate: gAutoUpdate=%d\n", gAutoUpdate);
}


static void set_updaterate(int rate)
{
	assert(rate != 0);
	gUpdateRate = rate;

	if(gAutoUpdate) {
		if(gTimerId != 0)
			gtk_timeout_remove(gTimerId);
		gTimerId = gtk_timeout_add(rate, &update_cb, 0);
	}
}

static void
handle_updaterate_cancel(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(widget);
}

static void
handle_updaterate_apply(GtkWidget *widget, gpointer data)
{
	set_updaterate(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gUpdateSpin)));
}

static void
handle_updaterate_ok(GtkWidget *widget, gpointer data)
{
	handle_updaterate_apply(widget, data);
	handle_updaterate_cancel(widget, data);
}

void
handle_updaterate(GtkWidget *widget, gpointer data)
{
	GtkWidget	*dialog,
			*okay,
			*cancel,
			*apply,
			*label;
	GtkAdjustment	*adj;
			
	dialog = gtk_dialog_new();
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	gtk_window_set_title(GTK_WINDOW(dialog), APPNAME": Update Rate" );

	label = gtk_label_new("Default: "DEF_UPDATE_STR" (milliseconds)");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, 1,1,1);

	adj = (GtkAdjustment*)gtk_adjustment_new(gUpdateRate, 
			30.0, 10000.0, 1.0, 1.0, 0.0);

	gUpdateSpin = gtk_spin_button_new(adj, 1.0, 1);

	gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(gUpdateSpin), FALSE);
	gtk_widget_set_usize(gUpdateSpin, 100, 0);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), gUpdateSpin, FALSE, TRUE, 0);


	okay = gtk_button_new_with_label("Okay");
	cancel = gtk_button_new_with_label("Cancel");
	apply = gtk_button_new_with_label("Apply");

	gtk_signal_connect_object(GTK_OBJECT(okay), "clicked", 
				  GTK_SIGNAL_FUNC(handle_updaterate_ok), 
				  (gpointer)dialog);
	gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", 
				  GTK_SIGNAL_FUNC(handle_updaterate_cancel), 
				  (gpointer)dialog);
	gtk_signal_connect_object(GTK_OBJECT(apply), "clicked", 
				  GTK_SIGNAL_FUNC(handle_updaterate_apply), 
				  (gpointer)dialog);

	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				cancel, 0,0,0); 
	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				okay, 0,0,0); 
	gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
				apply, 0,0,0); 



	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), gUpdateSpin);
	gtk_widget_show_all(dialog);
}


static void
update_dimension(int width, int height)
{
	static int screenWidth = 0;
	static int screenHeight = 0;

	
	if(!screenWidth || !screenHeight) {
		screenWidth = gdk_screen_width();
		screenHeight = gdk_screen_height();
	}

	while(width % 2 || height % 2) {
		if(width % 2)
			width--;
		if(height % 2)
			height--;
	}

	if(width > screenWidth)
		width = screenWidth;
	if(height > screenHeight)
		height = screenHeight;

	assert(width > 0);
	assert(height > 0);

	gWidth = width;
	gHeight = height;

	if(gZoomFactor <= 0)
		gZoomFactor = 1;

	gZWidth = gWidth / gZoomFactor;
	gZHeight = gHeight / gZoomFactor;

	if(gZWidth == 0) gZWidth = 1;
	if(gZHeight == 0) gZHeight = 1;

	gByteLine = ((gDepth == 24 ? 32 : gDepth) / 8) * gWidth;
	//gByteLine = (gDepth / 8) * gWidth;

	if(gCurX + gZWidth > screenWidth) {
		gCurX = (screenWidth - gZWidth);
	}

	if(gCurY + gZHeight > screenHeight) {
		gCurY = (screenHeight - gZHeight);
	}

	fprintf(stderr, "update_dimension: x=%d, y=%d, w=%d, h=%d, zw=%d, wh=%d\n",
			gCurX, gCurY, gWidth, gHeight, gZWidth, gZHeight);
}

static int 
handle_configure(GtkWidget *widget, GdkEventConfigure *event)
{
	int h;

	fprintf(stderr, "handle_configure: x=%d, y=%d, w=%d, h=%d\n",
			event->x, event->y, event->width, event->height);

	// update drawing coords
	h = event->height;
	h -= GTK_WIDGET(gMenu)->allocation.height;
	
	update_dimension(event->width, h);

	return TRUE;
}

static int 
handle_motion(GtkWidget *wid, GdkEventMotion *event)
{
	GdkModifierType state;
	int x, y;


	if(event->is_hint) {
		gdk_window_get_pointer (event->window, &x, &y, &state);
	} else {
		x = event->x;
		y = event->y;
		state = event->state;
	}
fprintf(stderr, "handle_motion: gCurX=%d, gCurY=%d\n",gCurX,gCurY);
fprintf(stderr, "handle_motion: x=%d, y=%d\n",x,y);

	if(state & GDK_BUTTON1_MASK) {
		if(!gIsDragging) {
			gIsDragging = 1;
			zoom_start_drag();
		}

		if(gFirstDrag) {
			gFirstDrag = 0;
		} else {
			zoom_draw_bands();
		}
		
		if(gIsDragging) {
			gCurX = event->x_root;
			gCurY = event->y_root;
			state = event->state;

			zoom_validate_position(&gCurX, &gCurY);
			zoom_draw_bands();
fprintf(stderr, "handle_motion: (validated)gCurX=%d, gCurY=%d\n",gCurX,gCurY);

			zoom_update_image();
		}
	}

	return TRUE;
}

static void 
handle_pressed(GtkWidget *wid, GdkEventButton *event)
{
	gtk_grab_add(wid); 
}

static void 
handle_released(GtkWidget *wid, GdkEventButton *event)
{
	gtk_grab_remove(wid);

	if(gIsDragging) {
		gIsDragging = 0;
		gFirstDrag = 1;
		zoom_draw_bands();
		zoom_end_drag();
	}
}

static void
handle_expose(GtkWidget *wid, GdkEventExpose *event)
{
	zoom_update_image();
}


static GtkWidget*
create_menu()
{
	GtkItemFactory *factory;
	GtkAccelGroup *agroup;
	static int items = sizeof(gMenuItems) / sizeof(gMenuItems[0]);


	agroup = gtk_accel_group_new();

	factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", agroup);

	gtk_item_factory_create_items(factory, items, gMenuItems, NULL);


	gtk_window_add_accel_group(GTK_WINDOW(gWindow), agroup);
	return gtk_item_factory_get_widget(factory, "<main>");
}

void 
gui_init()
{
	GtkWidget *box;

	box = gtk_vbox_new(FALSE, 0);

	gIsDragging = 0;

	gWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(gWindow), APPNAME);

	// quit when 'the big X' is pressed...
	gtk_signal_connect(GTK_OBJECT(gWindow), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_main_quit), NULL); 
	gtk_widget_set_usize(GTK_WIDGET(gWindow), DEFAULT_WIDTH, DEFAULT_HEIGHT);
	gtk_widget_realize(GTK_WIDGET(gWindow));

	gDrawing = gtk_drawing_area_new();

	gtk_signal_connect(GTK_OBJECT (gDrawing), "expose_event", 
			   (GtkSignalFunc)handle_expose, NULL); 
	gtk_signal_connect(GTK_OBJECT(gWindow),"configure_event", 
			   (GtkSignalFunc)handle_configure, NULL); 
	gtk_signal_connect(GTK_OBJECT(gDrawing), "motion_notify_event", 
			   (GtkSignalFunc)handle_motion, NULL); 
	gtk_signal_connect(GTK_OBJECT(gDrawing), "button_press_event",
			   (GtkSignalFunc)handle_pressed, NULL);
	gtk_signal_connect(GTK_OBJECT(gDrawing), "button_release_event",
			   (GtkSignalFunc)handle_released, NULL);

	gtk_widget_set_events(GTK_WIDGET(gDrawing), 
				GDK_EXPOSURE_MASK | 
				GDK_BUTTON_PRESS_MASK | 
				GDK_BUTTON_RELEASE_MASK |
				GDK_LEAVE_NOTIFY_MASK | 
				GDK_POINTER_MOTION_MASK |
				GDK_POINTER_MOTION_HINT_MASK);

	gMenu = create_menu();

	gtk_box_pack_start(GTK_BOX(box), gMenu, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box), gDrawing, TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(gWindow), box);
	//gtk_widget_show(gDrawing);
	gVisual = gdk_visual_get_system();
	gtk_widget_realize(GTK_WIDGET(box));
	gtk_widget_show(GTK_WIDGET(box));
	gtk_widget_realize(GTK_WIDGET(gDrawing));
	gtk_widget_show(GTK_WIDGET(gDrawing));

	gDepth = gVisual->depth;
	fprintf(stderr,"Server Depth: %d\n", gDepth);
	gByteLine = ((gDepth == 24 ? 32 : gDepth) / 8) * DEFAULT_WIDTH;
	//gByteLine = (gDepth / 8) * DEFAULT_WIDTH;
}


int 
main(int argc, char **argv)
{
	gtk_init(&argc, &argv);
	gui_init();
	zoom_init();

	gtk_widget_show_all(gWindow);

	gtk_main();
	return 0;
}
