// vim:ts=8:sw=8
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>

#include <X11/X.h>
#include <X11/Xlib.h>

#include <gdk_imlib.h>

#include "scale.h"

#ifndef DEBUG
#define fprintf(args...) do { }while(0)
#endif

extern int	gWidth,
		gHeight,
		gHWidth,
		gHHeight,
		gZWidth,
		gZHeight,
		gByteLine,
		gDepth,
		gCurX,
		gCurY;
extern int 		gFirstDrag; // i still remember mine...
extern int 		gIsDragging;

extern GtkWidget 	*gWindow;
extern GtkWidget 	*gDrawing;
extern GdkVisual	*gVisual;

static Display		*gxDisplay;
static GC		gxGC;
static GC		gxRootGC;
static Visual		*gxVisual;
static Window		gxDrawable;
static Window		gxRootWindow;
static XWindowAttributes gxRootAttr;

static XImage 		*gLastImage;

void zoom_init()
{
	GdkGC *tmpgc = gdk_gc_new(gDrawing->window);

	gxGC = GDK_GC_XGC(tmpgc);
	gxDrawable = GDK_WINDOW_XWINDOW(gDrawing->window),
	gxDisplay = GDK_GC_XDISPLAY(tmpgc);
	gxRootWindow = GDK_ROOT_WINDOW();
	gxVisual = GDK_VISUAL_XVISUAL(gVisual);
	XGetWindowAttributes(gxDisplay, gxRootWindow, &gxRootAttr);

	gxRootGC = XCreateGC(gxDisplay, gxDrawable, 0, NULL);

	gLastImage = 0;
}

static XImage*
zoom(XImage *img)
{
	XImage *nImage = NULL;

	unsigned char *nData = (unsigned char*)malloc(gByteLine * gHeight);

	//void RectStretch(long xs1,long ys1,long xs2,long ys2,
        //          long xd1,long yd1,long xd2,long yd2,
        //          unsigned char *aSrcImage, unsigned aSrcStride,
        //          unsigned char *aDstImage, unsigned aDstStride,
        //          unsigned aDepth);
	RectStretch(0,0, img->width-1, img->height-1,
		    0,0, (long)gWidth-1, (long)gHeight-1,
		    (unsigned char*)img->data, img->bytes_per_line,
		    nData, gByteLine,
		    gDepth);

	nImage = XCreateImage(gxDisplay,
				gxVisual,
				gDepth,
				ZPixmap,
				0,  //offset
				(char*)nData, //data
				gWidth, 
				gHeight,
				//gDepth,  // bitmap_pad
				img->bitmap_pad,
				gByteLine);  // bytes per line
	
	return nImage;
}

void zoom_update_image()
{
	XImage *screen;
	XImage *scaled;
	
	assert(gCurY >= 0);
	assert(gCurX >= 0);
	assert(gZWidth >= 0);
	assert(gZHeight >= 0);
	assert(gDepth >= 0);

	fprintf(stderr, "zoom_update_image: x=%d, y=%d, w=%d, h=%d\n",
				gCurX, gCurY, gZWidth, gZHeight);

	if(gLastImage)
		XDestroyImage(gLastImage);
			
	screen = XGetImage(gxDisplay, gxRootWindow,
			   gCurX, gCurY,
			   gZWidth, gZHeight,
			   AllPlanes, ZPixmap);

	if(screen == NULL) {
		fprintf(stderr,"XGetImage returns 0\n");
		return;
	}

	scaled = zoom(screen);

	assert(scaled != NULL);

	XPutImage(gxDisplay,
		  gxDrawable,
		  gxGC,
		  scaled,
		  0,0,
		  0,0,
		  gWidth, gHeight);
	
	gLastImage = scaled;
	XDestroyImage(screen);
	screen = 0;
}

void zoom_draw_bands()
{
	XDrawRectangle(gxDisplay, gxRootWindow, gxRootGC, 
			gCurX-1, gCurY-1, gZWidth+1, gZHeight+1);
}

void zoom_start_drag()
{
	/* set up for drawing a flashing rectangle */
	XSetFunction(gxDisplay, gxRootGC, GXinvert);
	XSetSubwindowMode(gxDisplay, gxRootGC, IncludeInferiors);
}

void zoom_end_drag()
{
	XUngrabPointer(gxDisplay, CurrentTime);
	XSetFunction(gxDisplay, gxRootGC, GXcopy);
	XSetSubwindowMode(gxDisplay, gxRootGC, ClipByChildren);
	XSetPlaneMask(gxDisplay, gxRootGC, AllPlanes);
	XSync(gxDisplay, False);
	
}

void zoom_validate_position()
{
	if(gCurX + gZWidth >= gxRootAttr.width)
		gCurX = gxRootAttr.width - gZWidth - 1;
	if(gCurY + gZHeight >= gxRootAttr.height)
		gCurY = gxRootAttr.height - gZHeight - 1;
}

void zoom_save(char *fname)
{
	GdkImlibImage *image;
	GdkImlibSaveInfo info;
	
	gdk_imlib_init();

	info.quality = 100;
	info.scaling = 0;
	info.xjustification = 0;
	info.yjustification = 0;

	image = gdk_imlib_create_image_from_data(gLastImage->data, NULL, 
						gLastImage->width,
						gLastImage->height);

	gdk_imlib_save_image(image, fname, &info);
	
#if 0
	FILE *fd;
	png_structp png_eng;
	png_infop png_inf;
	int i, r;
	png_bytep row;
	png_bytep rowbuff;

	if(!gLastImage) {
		fprintf(stderr, "zoom_save: Last image null\n");
		return;
	}
	
	fprintf(stderr, "Saving: %s\n", fname);

	fd = fopen(fname, "w");
	if(fd == 0) {
		fprintf(stderr,"Error opening file\n");
		return;
	}

	png_eng = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	png_inf = png_create_info_struct(png_eng);

	png_init_io(png_eng, fd);

	png_inf->width          = gLastImage->width;
	png_inf->height         = gLastImage->height;
	png_inf->bit_depth      = gLastImage->depth;
//	png_inf->gamma          = gLastImagegimp_gamma();
	png_inf->sig_bit.red    = 8;
	png_inf->sig_bit.green  = 8;
	png_inf->sig_bit.blue   = 8;
	png_inf->interlace_type = PNG_INTERLACE_NONE;
	//png_inf->valid          |= PNG_INFO_gAMA;
	png_inf->color_type = PNG_COLOR_TYPE_RGB;

	png_write_info(png_eng, png_inf);

	rowbuff = (png_bytep)malloc(gLastImage->width * 4);

	for(i = 0; i < gLastImage->height; i++) {
		row = gLastImage->data + (i * (gLastImage->bytes_per_line));
		// fill up buffer
		for(r = 0; r < gLastImage->width; r += 2) {
			rowbuff[r] = row[0];
			rowbuff[r] |= row[1];
		}
		
/*
		fprintf(stderr, "byte 1 on row %d, 0x%x\n", i, *row);
		fprintf(stderr, "byte 2 on row %d, 0x%x\n", i, *(row + 1));
		fprintf(stderr, "byte 3 on row %d, 0x%x\n", i, *(row + 2));
		fprintf(stderr, "byte 4 on row %d, 0x%x\n", i, *(row + 3));
		fprintf(stderr, "byte 5 on row %d, 0x%x\n", i, *(row + 4));
		fprintf(stderr, "byte 6 on row %d, 0x%x\n", i, *(row + 5));
*/
		png_write_row(png_eng, rowbuff);
	}

	png_write_end(png_eng, png_inf);
	png_write_destroy(png_eng);

	free(png_eng);
	free(png_inf);

	fclose(fd);
#endif
}	
